@@ -1,6 +1,7 @@
 /*
  *  Copyright (C) 2002 Jasper Huijsmans (huysmans@users.sourceforge.net)
  *  Copyright (C) 2003 Benedikt Meurer (benedikt.meurer@unix-ag.uni-siegen.de)
+ *  Copyright (c) 2004-2007 Brian Tarricone <bjt23@cornell.edu>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,6 +24,10 @@
 
 #include <stdio.h>
 
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
@@ -31,15 +36,28 @@
 #include <stdlib.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#endif
+
 #include <glib.h>
 #include <gdk/gdkx.h>
+#include <gtk/gtk.h>
+
 #include <libxfce4util/libxfce4util.h>
-#include <libxfcegui4/dialogs.h>
 
 #include "xfdesktop-common.h"
 
+#ifndef O_BINARY
+#define O_BINARY  0
+#endif
+
 gboolean
-is_backdrop_list(const gchar *path)
+xfdesktop_backdrop_list_is_valid(const gchar *path)
 {
     FILE *fp;
     gchar buf[512];
@@ -51,7 +69,7 @@ is_backdrop_list(const gchar *path)
     if(!(fp = fopen (path, "r")))
         return FALSE;
     
-    if(fgets(buf, size, fp) > 0 && !strncmp(LIST_TEXT, buf, size - 1))
+    if(fgets(buf, size, fp) && !strncmp(LIST_TEXT, buf, size - 1))
         is_list = TRUE;
     fclose(fp);
     
@@ -59,33 +77,180 @@ is_backdrop_list(const gchar *path)
 }
 
 gchar **
-get_list_from_file(const gchar *filename)
+xfdesktop_backdrop_list_load(const gchar *filename,
+                             gint *n_items,
+                             GError **error)
 {
-    gchar *contents;
-    GError *error = NULL;
-    gchar **files = NULL;
-    gsize length;
-    
-    files = NULL;
-    
-    if(!g_file_get_contents(filename, &contents, &length, &error)) {
-        xfce_err("Unable to get backdrop image list from file %s: %s",
-                filename, error->message);
-        g_error_free(error);
+    gchar *contents = NULL, **files = NULL, *p, *q;
+    gsize length = 0;
+    gint arr_size = 10, count = 0;
+
+    g_return_val_if_fail(filename && (!error || !*error), NULL);
+
+    if(!g_file_get_contents(filename, &contents, &length, error))
+        return NULL;
+
+    if(strncmp(LIST_TEXT, contents, sizeof(LIST_TEXT) - 1)) {
+        if(error) {
+            g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED,
+                        _("Backdrop list file is not valid"));
+        }
+        g_free(contents);
         return NULL;
     }
-    
-    if(strncmp(LIST_TEXT, contents, sizeof(LIST_TEXT) - 1)) {
-        xfce_err("Not a backdrop image list file: %s", filename);
-        goto finished;
+
+    /* i'd use g_strsplit() here, but then counting is slower.  we can
+     * also filter out blank lines */
+    files = g_malloc(sizeof(gchar *) * (arr_size+1));
+    p = contents + sizeof(LIST_TEXT);
+    while(p && *p) {
+        q = strstr(p, "\n");
+        if(q) {
+            if(p == q)  /* blank line */
+                continue;
+            *q = 0;
+        } else
+            q = contents + length;  /* assume no trailing '\n' at EOF */
+
+        if(count == arr_size) {
+            arr_size += 10;
+            files = g_realloc(files, sizeof(gchar *) * (arr_size+1));
+        }
+
+        files[count++] = g_strdup(p);
+        if(q != contents + length)
+            p = q + 1;
+    }
+    files[count] = NULL;
+    files = g_realloc(files, sizeof(gchar *) * (count+1));
+
+    if(n_items)
+        *n_items = count;
+
+    g_free(contents);
+
+    return files;
+}
+
+gboolean
+xfdesktop_backdrop_list_save(const gchar *filename,
+                             gchar * const *files,
+                             GError **error)
+{
+    gboolean ret = FALSE;
+    gchar *filename_new;
+    FILE *fp;
+    gint i;
+
+    g_return_val_if_fail(filename && (!error || !*error), FALSE);
+
+    filename_new = g_strconcat(filename, ".new", NULL);
+    fp = fopen(filename_new, "w");
+    if(fp) {
+        fprintf(fp, "%s\n", LIST_TEXT);
+        if(files) {
+            for(i = 0; files[i]; ++i)
+                fprintf(fp, "%s\n", files[i]);
+        }
+        if(!fclose(fp)) {
+            if(!rename(filename_new, filename))
+                ret = TRUE;
+            else  {
+                if(error) {
+                    g_set_error(error, G_FILE_ERROR,
+                                g_file_error_from_errno(errno),
+                                "%s", g_strerror(errno));
+                }
+                unlink(filename_new);
+            }
+        } else {
+            if(error) {
+                g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(errno),
+                            "%s", g_strerror(errno));
+            }
+            unlink(filename_new);
+        }
+    } else if(error) {
+        g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(errno),
+                    "%s", g_strerror(errno));
+    }
+
+    g_free(filename_new);
+
+    return ret;
+}
+
+gchar *
+xfdesktop_backdrop_list_choose_random(const gchar *filename,
+                                      GError **error)
+{
+    static gboolean __initialized = FALSE;
+    static gint previndex = -1;
+    gchar **files, *file = NULL;
+    gint n_items = 0, cur_file, i, tries = 0;
+
+    g_return_val_if_fail(filename && (!error || !*error), NULL);
+
+    files = xfdesktop_backdrop_list_load(filename, &n_items, error);
+    if(!files)
+        return NULL;
+    if(!n_items) {
+        if(error) {
+            g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED,
+                        _("Backdrop list file is not valid"));
+        }
+        g_strfreev(files);
+        return NULL;
+    }
+
+    if(1 == n_items) {
+        file = g_strdup(files[0]);
+        g_strfreev(files);
+        return file;
+    }
+
+    /* NOTE: 4.3BSD random()/srandom() are a) stronger and b) faster than
+    * ANSI-C rand()/srand(). So we use random() if available */
+    if(G_UNLIKELY(!__initialized))    {
+        guint seed = time(NULL) ^ (getpid() + (getpid() << 15));
+#ifdef HAVE_SRANDOM
+        srandom(seed);
+#else
+        srand(seed);
+#endif
+        __initialized = TRUE;
     }
     
-    files = g_strsplit(contents + sizeof(LIST_TEXT), "\n", -1);
-    
-    finished:
-    g_free(contents);
-    
-    return files;
+    do {
+        if(tries++ == n_items) {
+            /* this isn't precise, but if we've failed to get a good
+             * image after all this time, let's just give up */
+            g_warning("Unable to find good image from list; giving up");
+            g_strfreev(files);
+            return NULL;
+        }
+
+        do {
+#ifdef HAVE_SRANDOM
+            cur_file = random() % n_items;
+#else
+            cur_file = rand() % n_items;
+#endif
+        } while(cur_file == previndex && G_LIKELY(previndex != -1));
+
+    } while(!xfdesktop_image_file_is_valid(files[cur_file]));
+
+    previndex = cur_file;
+    file = files[cur_file];
+
+    /* don't use strfreev() and avoid an extra alloc */
+    for(i = 0; files[i]; ++i) {
+        if(i != cur_file)
+            g_free(files[i]);
+    }
+    g_free(files);
+
+    return file;
 }
 
 static void
@@ -99,101 +264,41 @@ pixbuf_loader_size_cb(GdkPixbufLoader *loader, gint width, gint height,
 }
 
 gboolean
-xfdesktop_check_image_file(const gchar *filename)
+xfdesktop_image_file_is_valid(const gchar *filename)
 {
     GdkPixbufLoader *loader;
-    FILE *fp;
+    int fd;
     gboolean size_read = FALSE;
     guchar buf[4096];
-    gint len;
+    gsize len;
+
+    g_return_val_if_fail(filename, FALSE);
     
-    fp = fopen(filename, "rb");
-    if(!fp)
+    fd = open(filename, O_RDONLY|O_BINARY);
+    if(fd < 0)
         return FALSE;
     
     loader = gdk_pixbuf_loader_new();
     g_signal_connect(G_OBJECT(loader), "size-prepared",
             G_CALLBACK(pixbuf_loader_size_cb), &size_read);
     
-    while(!feof(fp) && !ferror(fp)) {
-        if((len=fread(buf, 1, sizeof(buf), fp)) > 0) {
+    do {
+        len = read(fd, buf, sizeof(buf));
+        if(len > 0) {
             if(!gdk_pixbuf_loader_write(loader, buf, len, NULL))
                 break;
             if(size_read)
                 break;
         }
-    }
+    } while(len > 0);
     
-    fclose(fp);
+    close(fd);
     gdk_pixbuf_loader_close(loader, NULL);
     g_object_unref(G_OBJECT(loader));
     
     return size_read;
 }
 
-gchar *
-xfce_desktop_get_menufile(void)
-{
-    XfceKiosk *kiosk;
-    gboolean user_menu;
-    gchar filename[PATH_MAX], searchpath[PATH_MAX*3+2], **all_dirs;
-    const gchar *userhome = xfce_get_homedir();
-    gint i;
-
-    kiosk = xfce_kiosk_new("xfdesktop");
-    user_menu = xfce_kiosk_query(kiosk, "UserMenu");
-    xfce_kiosk_free(kiosk);
-    
-    if(!user_menu) {
-        all_dirs = xfce_resource_lookup_all(XFCE_RESOURCE_CONFIG,
-                                            "xfce4/desktop/");
-        
-        for(i = 0; all_dirs[i]; i++) {
-            if(strstr(all_dirs[i], userhome) != all_dirs[i]) {
-                g_snprintf(searchpath, PATH_MAX*3+2,
-                           "%s%%F.%%L:%s%%F.%%l:%s%%F",
-                           all_dirs[i], all_dirs[i], all_dirs[i]);
-                if(xfce_get_path_localized(filename, PATH_MAX, searchpath,
-                                           "menu.xml", G_FILE_TEST_IS_REGULAR))
-                {
-                    g_strfreev(all_dirs);
-                    return g_strdup(filename);
-                }
-            }            
-        }
-        g_strfreev(all_dirs);
-    } else {
-        gchar *menu_file = xfce_resource_save_location(XFCE_RESOURCE_CONFIG,
-                                                       "xfce4/desktop/menu.xml",
-                                                       FALSE);
-        if(menu_file && g_file_test(menu_file, G_FILE_TEST_IS_REGULAR))
-            return menu_file;
-        else if(menu_file)
-            g_free(menu_file);
-        
-        all_dirs = xfce_resource_lookup_all(XFCE_RESOURCE_CONFIG,
-                                            "xfce4/desktop/");
-        for(i = 0; all_dirs[i]; i++) {
-            if(strstr(all_dirs[i], userhome) != all_dirs[i]) {
-                g_snprintf(searchpath, PATH_MAX*3+2,
-                           "%s%%F.%%L:%s%%F.%%l:%s%%F",
-                           all_dirs[i], all_dirs[i], all_dirs[i]);
-                if(xfce_get_path_localized(filename, PATH_MAX, searchpath,
-                                           "menu.xml", G_FILE_TEST_IS_REGULAR))
-                {
-                    g_strfreev(all_dirs);
-                    return g_strdup(filename);
-                }        
-            }
-        }
-        g_strfreev(all_dirs);
-    }
-
-    g_warning("%s: Could not locate a menu definition file", PACKAGE);
-
-    return NULL;
-}
-
 gboolean
 xfdesktop_check_is_running(Window *xid)
 {
@@ -288,4 +393,3 @@ xfdesktop_popup_grab_available (GdkWindow *win, guint32 timestamp)
 
     return (!grab_failed);
 }
-
